from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from throttled.rate_limter import RateLimitResult


class BaseThrottledError(Exception):
    pass


class SetUpError(BaseThrottledError):
    pass


class DataError(BaseThrottledError):
    pass


class StoreUnavailableError(BaseThrottledError):
    pass


class LimitedError(BaseThrottledError):
    def __init__(self, rate_limit_result: Optional["RateLimitResult"] = None):
        self.rate_limit_result: Optional["RateLimitResult"] = rate_limit_result
        if not self.rate_limit_result or not self.rate_limit_result.state:
            message: str = "Rate limit exceeded"
        else:
            message: str = (
                "Rate limit exceeded: remaining={remaining}, "
                "reset_after={reset_after}"
            ).format(
                remaining=self.rate_limit_result.state.remaining,
                reset_after=self.rate_limit_result.state.reset_after,
            )
        super().__init__(message)



from enum import Enum

from .types import RateLimiterTypeT


class StoreType(Enum):
    REDIS: str = "redis"
    MEMORY: str = "memory"


class StoreTTLState(Enum):
    NOT_TTL: int = -1
    NOT_EXIST: int = -2


class RateLimiterType(Enum):
    FIXED_WINDOW: RateLimiterTypeT = "fixed_window"
    SLIDING_WINDOW: RateLimiterTypeT = "sliding_window"
    LEAKING_BUCKET: RateLimiterTypeT = "leaking_bucket"
    TOKEN_BUCKET: RateLimiterTypeT = "token_bucket"
    GCRA: RateLimiterTypeT = "gcra"



import math
import time
from enum import Enum
from typing import List, Optional, Sequence, Tuple, Type

from redis.commands.core import Script

from ..constants import RateLimiterType, StoreType
from ..store import BaseAtomicAction, MemoryStoreBackend, RedisStoreBackend
from ..types import AtomicActionTypeT, KeyT, RateLimiterTypeT, StoreValueT
from . import BaseRateLimiter, RateLimitResult, RateLimitState


class GCRAAtomicActionType(Enum):
    """Enumeration for types of AtomicActions used in GCRARateLimiter."""

    PEEK: AtomicActionTypeT = "peek"
    LIMIT: AtomicActionTypeT = "limit"


class RedisLimitAtomicAction(BaseAtomicAction):
    """Redis-based implementation of AtomicAction for GCRARateLimiter's limit operation.
    Inspire by [Rate Limiting, Cells, and GCRA](https://brandur.org/rate-limiting).
    """

    TYPE: AtomicActionTypeT = GCRAAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.REDIS.value

    SCRIPTS: str = """
    local emission_interval = tonumber(ARGV[1])
    local capacity = tonumber(ARGV[2])
    local cost = tonumber(ARGV[3])

    local jan_1_2025 = 1735660800
    local now = redis.call("TIME")
    now = (now[1] - jan_1_2025) + (now[2] / 1000000)

    local last_tat = redis.call("GET", KEYS[1])
    if not last_tat then
        last_tat = now
    else
        last_tat = tonumber(last_tat)
    end

    -- Calculate the fill time required for the current cost.
    local fill_time_for_cost = cost * emission_interval
    -- Calculate the fill time required for the full capacity.
    local fill_time_for_capacity = capacity * emission_interval
    -- Calculate the theoretical arrival time (TAT) for the current request.
    local tat = math.max(now, last_tat) + fill_time_for_cost
    -- Calculate the the time when the request would be allowed.
    local allow_at = tat - fill_time_for_capacity
    -- Calculate the time elapsed since the request would be allowed.
    local time_elapsed = now - allow_at


    local limited = 0
    local retry_after = 0
    local reset_after = tat - now
    local remaining = math.floor(time_elapsed / emission_interval)
    if remaining < 0 then
        limited = 1
        retry_after = time_elapsed * -1
        reset_after = math.max(0, last_tat - now)
        remaining = math.min(capacity, cost + remaining)
    else
        redis.call("SET", KEYS[1], tat, "EX", math.ceil(reset_after))
    end

    -- use tostring to avoid lost precision.
    return {limited, remaining, tostring(reset_after), tostring(retry_after)}
    """

    def __init__(self, backend: RedisStoreBackend):
        self._script: Script = backend.get_client().register_script(self.SCRIPTS)

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int, float, float]:
        limited, remaining, reset_after, retry_after = self._script(keys, args)
        return limited, remaining, float(reset_after), float(retry_after)


class RedisPeekAtomicAction(RedisLimitAtomicAction):
    """
    Redis-based implementation of AtomicAction for GCRARateLimiter's peek operation.
    """

    TYPE: AtomicActionTypeT = GCRAAtomicActionType.PEEK.value

    SCRIPTS: str = """
    local emission_interval = tonumber(ARGV[1])
    local capacity = tonumber(ARGV[2])

    local jan_1_2025 = 1735660800
    local now = redis.call("TIME")
    now = (now[1] - jan_1_2025) + (now[2] / 1000000)

    local tat = redis.call("GET", KEYS[1])
    if not tat then
        tat = now
    else
        tat= tonumber(tat)
    end

    -- Calculate the fill time required for the full capacity.
    local fill_time_for_capacity = capacity * emission_interval
    -- Calculate the the time when the request would be allowed.
    local allow_at = math.max(tat, now) - fill_time_for_capacity
    -- Calculate the time elapsed since the request would be allowed.
    local time_elapsed = now - allow_at

    local limited = 0
    local retry_after = 0
    local reset_after = math.max(0, tat - now)
    local remaining = math.floor(time_elapsed / emission_interval)
    if remaining < 1 then
        limited = 1
        remaining = 0
        retry_after = time_elapsed * -1
    end

    -- use tostring to avoid lost precision.
    return {limited, remaining, tostring(reset_after), tostring(retry_after)}
    """


class MemoryLimitAtomicAction(BaseAtomicAction):
    """Memory-based implementation of AtomicAction for GCRARateLimiter's limit operation.
    Inspire by [Rate Limiting, Cells, and GCRA](https://brandur.org/rate-limiting).
    """

    TYPE: AtomicActionTypeT = GCRAAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.MEMORY.value

    def __init__(self, backend: MemoryStoreBackend):
        self._backend: MemoryStoreBackend = backend

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int, float, float]:
        with self._backend.lock:
            key: str = keys[0]
            emission_interval: float = args[0]
            capacity: int = args[1]
            cost: int = args[2]

            now: float = time.time()
            last_tat: float = self._backend.get(key) or now

            fill_time_for_cost: float = cost * emission_interval
            fill_time_for_capacity: float = capacity * emission_interval
            tat: float = max(now, last_tat) + fill_time_for_cost
            allow_at: float = tat - fill_time_for_capacity
            time_elapsed: float = now - allow_at

            remaining: int = math.floor(time_elapsed / emission_interval)
            if remaining < 0:
                limited: int = 1
                retry_after: float = -time_elapsed
                reset_after: float = max(0, last_tat - now)
                remaining: int = min(capacity, cost + remaining)
            else:
                limited: int = 0
                retry_after: float = 0
                reset_after: float = tat - now
                self._backend.set(key, tat, math.ceil(reset_after))

        return limited, remaining, reset_after, retry_after


class MemoryPeekAtomicAction(MemoryLimitAtomicAction):
    """
    Memory-based implementation of AtomicAction for GCRARateLimiter's peek operation.
    """

    TYPE: AtomicActionTypeT = GCRAAtomicActionType.PEEK.value

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int, float, float]:
        with self._backend.lock:
            key: str = keys[0]
            emission_interval: float = args[0]
            capacity: int = args[1]

            now: float = time.time()
            tat: float = self._backend.get(key) or now
            fill_time_for_capacity: float = capacity * emission_interval
            allow_at: float = max(now, tat) - fill_time_for_capacity
            time_elapsed: float = now - allow_at

            reset_after: float = max(0, tat - now)
            remaining: int = math.floor(time_elapsed / emission_interval)
            if remaining < 1:
                limited: int = 1
                remaining: int = 0
                retry_after: float = -time_elapsed
            else:
                limited: int = 0
                retry_after: float = 0
        return limited, remaining, reset_after, retry_after


class GCRARateLimiter(BaseRateLimiter):
    """Concrete implementation of BaseRateLimiter using GCRA as algorithm."""

    class Meta:
        type: RateLimiterTypeT = RateLimiterType.GCRA.value

    @classmethod
    def _default_atomic_action_classes(cls) -> List[Type[BaseAtomicAction]]:
        return [
            RedisPeekAtomicAction,
            RedisLimitAtomicAction,
            MemoryLimitAtomicAction,
            MemoryPeekAtomicAction,
        ]

    @classmethod
    def _supported_atomic_action_types(cls) -> List[AtomicActionTypeT]:
        return [GCRAAtomicActionType.LIMIT.value, GCRAAtomicActionType.PEEK.value]

    def _prepare(self, key: str) -> Tuple[str, float, int]:
        emission_interval: float = self.quota.get_period_sec() / self.quota.get_limit()
        return key, emission_interval, self.quota.burst

    def _limit(self, key: str, cost: int = 1) -> RateLimitResult:
        formatted_key, emission_interval, capacity = self._prepare(key)
        limited, remaining, reset_after, __ = self._atomic_actions[
            GCRAAtomicActionType.LIMIT.value
        ].do([formatted_key], [emission_interval, capacity, cost])

        return RateLimitResult(
            limited=bool(limited),
            state=RateLimitState(
                limit=capacity, remaining=remaining, reset_after=reset_after
            ),
        )

    def _peek(self, key: str) -> RateLimitState:
        formatted_key, emission_interval, capacity = self._prepare(key)
        limited, remaining, reset_after, __ = self._atomic_actions[
            GCRAAtomicActionType.PEEK.value
        ].do([formatted_key], [emission_interval, capacity])
        return RateLimitState(
            limit=capacity, remaining=remaining, reset_after=reset_after
        )



import threading
from typing import Callable, Optional, Type

from .constants import RateLimiterType
from .exceptions import DataError, LimitedError
from .rate_limter import (
    BaseRateLimiter,
    Quota,
    RateLimiterRegistry,
    RateLimitResult,
    RateLimitState,
    per_min,
)
from .store import BaseStore, MemoryStore
from .types import KeyT, RateLimiterTypeT


class Throttled:
    def __init__(
        self,
        key: Optional[KeyT] = None,
        using: Optional[RateLimiterTypeT] = None,
        quota: Optional[Quota] = None,
        store: Optional[BaseStore] = None,
    ):
        """Initializes the Throttled class.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :param using: The type of rate limiter to use, default: token_bucket.
        :param quota: The quota for the rate limiter, default: 60 requests per minute.
        :param store: The store to use for the rate limiter, default: MemoryStore.
        """
        # TODO Support key prefix.
        # TODO Support extract key from params.
        # TODO Support get cost weight by key.
        self.key: Optional[str] = key
        self._quota: Quota = quota or per_min(60)
        self._store: BaseStore = store or MemoryStore()
        self._limiter_cls: Type[BaseRateLimiter] = RateLimiterRegistry.get(
            using or RateLimiterType.TOKEN_BUCKET.value
        )

        self._lock: threading.Lock = threading.Lock()
        self._limiter: Optional[BaseRateLimiter] = None

    def _get_limiter(self) -> BaseRateLimiter:
        """Lazily initializes and returns the rate limiter instance."""
        if self._limiter:
            return self._limiter

        with self._lock:
            # Double-check locking to ensure thread safety.
            if self._limiter:
                return self._limiter

            self._limiter = self._limiter_cls(self._quota, self._store)
            return self._limiter

    def __call__(self, func: Callable) -> Callable:
        """Decorator to apply rate limiting to a function."""

        if not self.key:
            raise DataError("Decorator requires a non-empty key.")

        def _inner(*args, **kwargs):
            # TODO Add options to ignore state.
            result: RateLimitResult = self.limit(self.key)
            if result.limited:
                raise LimitedError(rate_limit_result=result)
            return func(*args, **kwargs)

        return _inner

    @classmethod
    def _validate_cost(cls, cost: int) -> None:
        """Validate the cost of the current request.
        :param cost: The cost of the current request in terms of
                     how much of the rate limit quota it consumes.
        :raise: DataError if the cost is not a positive integer.
        """
        if isinstance(cost, int) and cost > 0:
            return

        raise DataError(
            "Invalid cost: {cost}, Must be an integer greater "
            "than 0.".format(cost=cost)
        )

    def limit(self, key: KeyT, cost: int = 1) -> RateLimitResult:
        """Apply rate limiting logic to a given key with a specified cost.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :param cost: The cost of the current request in terms of how much
                     of the rate limit quota it consumes.
        :return: RateLimitResult: The result of the rate limiting check.
        :raise: DataError
        """
        self._validate_cost(cost)
        return self._get_limiter().limit(key, cost)

    def peek(self, key: KeyT) -> RateLimitState:
        """Retrieve the current state of rate limiter for the given key
           without actually modifying the state.
        :param key: The unique identifier for the rate limit subject.
                    eg: user ID or IP address.
        :return: RateLimitState - Representing the current state of
                 the rate limiter for the given key.
        """
        return self._get_limiter().peek(key)



from .constants import RateLimiterType
from .rate_limter import (
    BaseRateLimiter,
    Quota,
    Rate,
    RateLimiterMeta,
    RateLimiterRegistry,
    RateLimitResult,
    RateLimitState,
    per_day,
    per_hour,
    per_min,
    per_sec,
)
from .store import (
    BaseAtomicAction,
    BaseStore,
    BaseStoreBackend,
    MemoryStore,
    MemoryStoreBackend,
    RedisStore,
    RedisStoreBackend,
)
from .throttled import Throttled

__all__ = [
    # public module
    "exceptions",
    "constants",
    "types",
    "utils",
    # rate_limiter
    "per_sec",
    "per_min",
    "per_hour",
    "per_day",
    "Rate",
    "Quota",
    "RateLimitState",
    "RateLimitResult",
    "RateLimiterRegistry",
    "RateLimiterMeta",
    "BaseRateLimiter",
    # store
    "BaseStoreBackend",
    "BaseAtomicAction",
    "BaseStore",
    "MemoryStoreBackend",
    "MemoryStore",
    "RedisStoreBackend",
    "RedisStore",
    # throttled
    "Throttled",
    # constants
    "RateLimiterType",
]



import math
from enum import Enum
from typing import List, Optional, Sequence, Tuple, Type

from redis.commands.core import Script

from ..constants import RateLimiterType, StoreType
from ..store import BaseAtomicAction, MemoryStoreBackend, RedisStoreBackend
from ..types import (
    AtomicActionTypeT,
    KeyT,
    RateLimiterTypeT,
    StoreDictValueT,
    StoreValueT,
)
from ..utils import now_sec
from . import BaseRateLimiter, RateLimitResult, RateLimitState


class TokenBucketAtomicActionType(Enum):
    """Enumeration for types of AtomicActions used in TokenBucketRateLimiter."""

    LIMIT: AtomicActionTypeT = "limit"


class RedisLimitAtomicAction(BaseAtomicAction):
    """Redis-based implementation of AtomicAction for TokenBucketRateLimiter."""

    TYPE: AtomicActionTypeT = TokenBucketAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.REDIS.value

    SCRIPTS: str = """
    local rate = tonumber(ARGV[1])
    local capacity = tonumber(ARGV[2])
    local cost = tonumber(ARGV[3])
    local now = tonumber(ARGV[4])

    local expired = false
    local last_tokens = capacity
    local last_refreshed = now
    local bucket = redis.call("HMGET", KEYS[1], "tokens", "last_refreshed")

    if bucket[1] ~= false then
        expired = true
        last_tokens = tonumber(bucket[1])
        last_refreshed = tonumber(bucket[2])
    end

    local time_elapsed = math.max(0, now - last_refreshed)
    local tokens = math.min(capacity, last_tokens + (math.floor(time_elapsed * rate)))

    local limited = cost > tokens
    if not limited then
        tokens = tokens - cost
    end

    redis.call("HSET", KEYS[1], "tokens", tokens, "last_refreshed", now)

    local fill_time = capacity / rate
    if (expired or now - last_refreshed >= 1) then
        redis.call("EXPIRE", KEYS[1], math.ceil(2 * fill_time))
    end
    return {limited, tokens}
    """

    def __init__(self, backend: RedisStoreBackend):
        self._script: Script = backend.get_client().register_script(self.SCRIPTS)

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int]:
        return self._script(keys, args)


class MemoryLimitAtomicAction(BaseAtomicAction):
    """Memory-based implementation of AtomicAction for TokenBucketRateLimiter."""

    TYPE: AtomicActionTypeT = TokenBucketAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.MEMORY.value

    def __init__(self, backend: MemoryStoreBackend):
        self._backend: MemoryStoreBackend = backend

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int]:
        with self._backend.lock:
            key: str = keys[0]
            rate: float = args[0]
            capacity: int = args[1]
            cost: int = args[2]
            now: int = args[3]

            bucket: StoreDictValueT = self._backend.hgetall(key)
            last_tokens: int = bucket.get("tokens", capacity)
            last_refreshed: int = bucket.get("last_refreshed", now)

            time_elapsed: int = max(0, now - last_refreshed)
            tokens: int = min(capacity, last_tokens + (math.floor(time_elapsed * rate)))

            limited: int = (1, 0)[tokens >= cost]
            if not limited:
                tokens -= cost

            self._backend.hset(key, mapping={"tokens": tokens, "last_refreshed": now})

            fill_time: float = capacity / rate
            self._backend.expire(key, math.ceil(2 * fill_time))

            return limited, tokens


class TokenBucketRateLimiter(BaseRateLimiter):
    """Concrete implementation of BaseRateLimiter using token bucket as algorithm."""

    class Meta:
        type: RateLimiterTypeT = RateLimiterType.TOKEN_BUCKET.value

    @classmethod
    def _default_atomic_action_classes(cls) -> List[Type[BaseAtomicAction]]:
        return [RedisLimitAtomicAction, MemoryLimitAtomicAction]

    @classmethod
    def _supported_atomic_action_types(cls) -> List[AtomicActionTypeT]:
        return [TokenBucketAtomicActionType.LIMIT.value]

    def _prepare(self, key: str) -> Tuple[str, float, int]:
        rate: float = self.quota.get_limit() / self.quota.get_period_sec()
        return key, rate, self.quota.burst

    def _limit(self, key: str, cost: int = 1) -> RateLimitResult:
        formatted_key, rate, capacity = self._prepare(key)
        limited, tokens = self._atomic_actions[
            TokenBucketAtomicActionType.LIMIT.value
        ].do([formatted_key], [rate, capacity, cost, now_sec()])
        reset_after: int = math.ceil((capacity - tokens) / rate)

        return RateLimitResult(
            limited=bool(limited),
            state=RateLimitState(
                limit=capacity, remaining=tokens, reset_after=reset_after
            ),
        )

    def _peek(self, key: str) -> RateLimitState:
        now: int = now_sec()
        formatted_key, rate, capacity = self._prepare(key)

        bucket: StoreDictValueT = self._store.hgetall(formatted_key)
        last_tokens: int = bucket.get("tokens", capacity)
        last_refreshed: int = bucket.get("last_refreshed", now)

        time_elapsed: int = max(0, now - last_refreshed)
        tokens: int = min(capacity, last_tokens + (math.floor(time_elapsed * rate)))
        reset_after: int = math.ceil((capacity - tokens) / rate)

        return RateLimitState(limit=capacity, remaining=tokens, reset_after=reset_after)



from .base import (
    BaseRateLimiter,
    Quota,
    Rate,
    RateLimiterMeta,
    RateLimiterRegistry,
    RateLimitResult,
    RateLimitState,
    per_day,
    per_hour,
    per_min,
    per_sec,
)

# Trigger to register RateLimiter
from .fixed_window import FixedWindowRateLimiter
from .gcra import GCRARateLimiter
from .leaking_bucket import LeakingBucketRateLimiter
from .sliding_window import SlidingWindowRateLimiter
from .token_bucket import TokenBucketRateLimiter

__all__ = [
    "per_sec",
    "per_min",
    "per_hour",
    "per_day",
    "Rate",
    "Quota",
    "RateLimitState",
    "RateLimitResult",
    "RateLimiterRegistry",
    "RateLimiterMeta",
    "BaseRateLimiter",
    "FixedWindowRateLimiter",
    "SlidingWindowRateLimiter",
    "TokenBucketRateLimiter",
    "LeakingBucketRateLimiter",
    "GCRARateLimiter",
]



from enum import Enum
from typing import List, Optional, Sequence, Tuple, Type

from redis.commands.core import Script

from ..constants import RateLimiterType, StoreType
from ..store import BaseAtomicAction, MemoryStoreBackend, RedisStoreBackend
from ..types import AtomicActionTypeT, KeyT, RateLimiterTypeT, StoreValueT
from ..utils import now_sec
from . import BaseRateLimiter, RateLimitResult, RateLimitState


class FixedWindowAtomicActionType(Enum):
    """Enumeration for types of AtomicActions used in FixedWindowRateLimiter."""

    LIMIT: AtomicActionTypeT = "limit"


class RedisLimitAtomicAction(BaseAtomicAction):
    """Redis-based implementation of AtomicAction for FixedWindowRateLimiter."""

    TYPE: AtomicActionTypeT = FixedWindowAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.REDIS.value

    SCRIPTS: str = """
    local period = tonumber(ARGV[1])
    local limit = tonumber(ARGV[2])
    local cost = tonumber(ARGV[3])
    local current = redis.call("INCRBY", KEYS[1], cost)

    if current == cost then
        redis.call("EXPIRE", KEYS[1], period)
    end

    return {current > limit and 1 or 0, current}
    """

    def __init__(self, backend: RedisStoreBackend):
        self._script: Script = backend.get_client().register_script(self.SCRIPTS)

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int]:
        return self._script(keys, args)


class MemoryLimitAtomicAction(BaseAtomicAction):
    """Memory-based implementation of AtomicAction for FixedWindowRateLimiter."""

    TYPE: AtomicActionTypeT = FixedWindowAtomicActionType.LIMIT.value
    STORE_TYPE: str = StoreType.MEMORY.value

    def __init__(self, backend: MemoryStoreBackend):
        self._backend: MemoryStoreBackend = backend

    def do(
        self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]
    ) -> Tuple[int, int]:
        with self._backend.lock:
            key: str = keys[0]
            period: int = args[0]
            limit: int = args[1]
            cost: int = args[2]

            current: Optional[int] = self._backend.get(key)
            if current is None:
                current = cost
                self._backend.set(key, current, period)
            else:
                current += cost
                self._backend.get_client()[key] = current

            return (0, 1)[current > limit], current


class FixedWindowRateLimiter(BaseRateLimiter):
    """Concrete implementation of BaseRateLimiter using fixed window as algorithm."""

    class Meta:
        type: RateLimiterTypeT = RateLimiterType.FIXED_WINDOW.value

    @classmethod
    def _default_atomic_action_classes(cls) -> List[Type[BaseAtomicAction]]:
        return [RedisLimitAtomicAction, MemoryLimitAtomicAction]

    @classmethod
    def _supported_atomic_action_types(cls) -> List[AtomicActionTypeT]:
        return [FixedWindowAtomicActionType.LIMIT.value]

    def _prepare(self, key: str) -> Tuple[str, int, int]:
        period: int = self.quota.get_period_sec()
        period_key: str = f"{key}:period:{now_sec() // period}"
        return period_key, period, self.quota.get_limit()

    def _limit(self, key: str, cost: int = 1) -> RateLimitResult:
        period_key, period, limit = self._prepare(key)
        limited, current = self._atomic_actions[
            FixedWindowAtomicActionType.LIMIT.value
        ].do([period_key], [period, limit, cost])
        return RateLimitResult(
            limited=bool(limited),
            state=RateLimitState(
                limit=limit, remaining=max(0, limit - current), reset_after=period
            ),
        )

    def _peek(self, key: str) -> RateLimitState:
        period_key, period, limit = self._prepare(key)
        current: int = int(self._store.get(period_key) or 0)
        return RateLimitState(
            limit=limit, remaining=max(0, limit - current), reset_after=period
        )



import threading
from collections import OrderedDict
from typing import Any, Dict, Optional
from typing import OrderedDict as OrderedDictT
from typing import Type

from ..constants import StoreTTLState, StoreType
from ..exceptions import DataError, SetUpError
from ..types import KeyT, StoreBucketValueT, StoreDictValueT, StoreValueT
from ..utils import now_sec
from .base import BaseAtomicAction, BaseStore, BaseStoreBackend


class MemoryStoreBackend(BaseStoreBackend):
    """Backend for Memory Store."""

    def __init__(
        self, server: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ):
        super().__init__(server, options)

        max_size: Any = self.options.get("MAX_SIZE", 1024)
        if not (isinstance(max_size, int) and max_size > 0):
            raise SetUpError("MAX_SIZE must be a positive integer")

        self.max_size: int = max_size
        self.expire_info: Dict[str, float] = {}
        self.lock: threading.RLock = threading.RLock()
        self._client: OrderedDictT[KeyT, StoreBucketValueT] = OrderedDict()

    def get_client(self) -> OrderedDictT[KeyT, StoreBucketValueT]:
        return self._client

    def exists(self, key: KeyT) -> bool:
        return key in self._client

    def has_expired(self, key: KeyT) -> bool:
        return self.ttl(key) == StoreTTLState.NOT_EXIST.value

    def ttl(self, key: KeyT) -> int:
        exp: Optional[float] = self.expire_info.get(key)
        if exp is None:
            if not self.exists(key):
                return StoreTTLState.NOT_EXIST.value
            return StoreTTLState.NOT_TTL.value

        ttl: int = int(exp) - now_sec()
        if ttl <= 0:
            return StoreTTLState.NOT_EXIST.value
        return ttl

    def check_and_evict(self, key: KeyT) -> None:
        is_full: bool = len(self._client) >= self.max_size
        if is_full and not self.exists(key):
            pop_key, __ = self._client.popitem(last=False)
            self.expire_info.pop(pop_key, None)

    def expire(self, key: KeyT, timeout: int) -> None:
        self.expire_info[key] = now_sec() + timeout

    def get(self, key: KeyT) -> Optional[StoreValueT]:
        if self.has_expired(key):
            self.delete(key)
            return None

        value: Optional[StoreValueT] = self._client.get(key)
        if value is not None:
            self._client.move_to_end(key)
        return value

    def set(self, key: KeyT, value: StoreValueT, timeout: int) -> None:
        self.check_and_evict(key)
        self._client[key] = value
        self._client.move_to_end(key)
        self.expire(key, timeout)

    def hset(
        self,
        name: KeyT,
        key: Optional[KeyT] = None,
        value: Optional[StoreValueT] = None,
        mapping: Optional[StoreDictValueT] = None,
    ) -> None:
        if key is None and not mapping:
            raise DataError("hset must with key value pairs")

        kv: StoreDictValueT = {}
        if key is not None:
            kv[key] = value
        if mapping:
            kv.update(mapping)

        origin: Optional[StoreBucketValueT] = self._client.get(name)
        if origin is not None:
            if not isinstance(origin, dict):
                raise DataError("origin must be a dict")
            origin.update(kv)
        else:
            self.check_and_evict(key)
            self._client[name] = kv

        self._client.move_to_end(name)

    def hgetall(self, name: KeyT) -> StoreDictValueT:
        if self.has_expired(name):
            self.delete(name)
            return {}

        kv: Optional[StoreBucketValueT] = self._client.get(name)
        if not (kv is None or isinstance(kv, dict)):
            raise DataError("NumberLike value does not support hgetall")

        if kv is not None:
            self._client.move_to_end(name)

        return kv or {}

    def delete(self, key: KeyT) -> bool:
        try:
            self.expire_info.pop(key, None)
            del self._client[key]
        except KeyError:
            return False
        return True


class MemoryStore(BaseStore):
    """Concrete implementation of BaseStore using Memory as backend.

    Below are the performance benchmarks for different configurations of the LRU cache,
    tested using LeetCode problems (https://leetcode.cn/problems/lru-cache/):
    - LRU with Lock and Expiry  -> 265 ms, 76.8 MB
    - LRU with Lock only        -> 211 ms, 76.8 MB
    - LRU only                  -> 103 ms, 76.8 MB  (Beat 92.77% of submissions)
    - LRU implemented in Golang -> 86 ms,  76.43 MB (Beat 52.98% of submissions)
    """

    TYPE: str = StoreType.MEMORY.value

    def __init__(
        self, server: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ):
        self._backend: MemoryStoreBackend = MemoryStoreBackend(server, options)

    def exists(self, key: KeyT) -> bool:
        return self._backend.exists(key)

    def ttl(self, key: KeyT) -> int:
        return self._backend.ttl(key)

    def expire(self, key: KeyT, timeout: int) -> None:
        self._validate_timeout(timeout)
        self._backend.expire(key, timeout)

    def set(self, key: KeyT, value: StoreValueT, timeout: int) -> None:
        self._validate_timeout(timeout)
        with self._backend.lock:
            self._backend.set(key, value, timeout)

    def get(self, key: KeyT) -> Optional[StoreValueT]:
        with self._backend.lock:
            return self._backend.get(key)

    def hset(
        self,
        name: KeyT,
        key: Optional[KeyT] = None,
        value: Optional[StoreValueT] = None,
        mapping: Optional[StoreDictValueT] = None,
    ) -> None:
        with self._backend.lock:
            self._backend.hset(name, key, value, mapping)

    def hgetall(self, name: KeyT) -> StoreDictValueT:
        with self._backend.lock:
            return self._backend.hgetall(name)

    def make_atomic(self, action_cls: Type[BaseAtomicAction]) -> BaseAtomicAction:
        return action_cls(backend=self._backend)



from .base import BaseAtomicAction, BaseStore, BaseStoreBackend
from .memory import MemoryStore, MemoryStoreBackend
from .redis import RedisStore, RedisStoreBackend
from .redis_pool import (
    BaseConnectionFactory,
    ConnectionFactory,
    SentinelConnectionFactory,
    get_connection_factory,
)

__all__ = [
    "BaseStoreBackend",
    "BaseAtomicAction",
    "BaseStore",
    "MemoryStoreBackend",
    "MemoryStore",
    "RedisStoreBackend",
    "RedisStore",
    "BaseConnectionFactory",
    "ConnectionFactory",
    "SentinelConnectionFactory",
    "get_connection_factory",
]



from typing import Any, Dict, Optional, Type, Union

from redis import Redis

from ..constants import StoreType
from ..exceptions import DataError
from ..types import KeyT, StoreDictValueT, StoreValueT
from .base import BaseAtomicAction, BaseStore, BaseStoreBackend
from .redis_pool import BaseConnectionFactory, get_connection_factory


class RedisStoreBackend(BaseStoreBackend):
    """Backend for Redis store."""

    def __init__(
        self, server: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ):
        super().__init__(server, options)

        self._client: Optional[Redis] = None

        connection_factory_cls_path: Optional[str] = self.options.get(
            "CONNECTION_FACTORY_CLASS"
        )
        self._connection_factory: BaseConnectionFactory = get_connection_factory(
            connection_factory_cls_path, self.options
        )

    def get_client(self) -> Redis:
        if self._client is None:
            self._client = self._connection_factory.connect(self.server)
        return self._client


class RedisStore(BaseStore):
    """Concrete implementation of BaseStore using Redis as backend."""

    TYPE: str = StoreType.REDIS.value

    def __init__(
        self, server: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ):
        self._backend: RedisStoreBackend = RedisStoreBackend(server, options)

    def exists(self, key: KeyT) -> bool:
        return bool(self._backend.get_client().exists(key))

    def ttl(self, key: KeyT) -> int:
        return int(self._backend.get_client().ttl(key))

    def expire(self, key: KeyT, timeout: int) -> None:
        self._validate_timeout(timeout)
        self._backend.get_client().expire(key, timeout)

    def set(self, key: KeyT, value: StoreValueT, timeout: int) -> None:
        self._validate_timeout(timeout)
        self._backend.get_client().set(key, value, ex=timeout)

    @classmethod
    def _format_value(cls, value: StoreValueT) -> StoreValueT:
        float_value: float = float(value)
        if float_value.is_integer():
            return int(float_value)
        return float_value

    @classmethod
    def _format_key(cls, key: Union[bytes, str]) -> KeyT:
        if isinstance(key, bytes):
            return key.decode("utf-8")
        return key

    def get(self, key: KeyT) -> Optional[StoreValueT]:
        value: Optional[StoreValueT] = self._backend.get_client().get(key)
        if value is None:
            return None

        return self._format_value(value)

    def hset(
        self,
        name: KeyT,
        key: Optional[KeyT] = None,
        value: Optional[StoreValueT] = None,
        mapping: Optional[StoreDictValueT] = None,
    ) -> None:
        if key is None and not mapping:
            raise DataError("hset must with key value pairs")
        self._backend.get_client().hset(name, key, value, mapping)

    def hgetall(self, name: KeyT) -> StoreDictValueT:
        kv: Dict[KeyT, Optional[StoreValueT]] = self._backend.get_client().hgetall(name)
        return {self._format_key(k): self._format_value(v) for k, v in kv.items()}

    def make_atomic(self, action_cls: Type[BaseAtomicAction]) -> BaseAtomicAction:
        return action_cls(backend=self._backend)



import abc
from typing import Any, Dict, Optional, Sequence, Type

from ..exceptions import DataError, SetUpError
from ..types import KeyT, StoreDictValueT, StoreValueT


class BaseStoreBackend(abc.ABC):
    """Abstract class for all store backends."""

    def __init__(
        self, server: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ):
        self.server: Optional[str] = server
        self.options: Dict[str, Any] = options or {}

    @abc.abstractmethod
    def get_client(self) -> Any:
        raise NotImplementedError


class BaseAtomicAction(abc.ABC):
    """Abstract class for all atomic actions performed by a store backend."""

    # TYPE is the identifier of AtomicAction, must be unique under STORE_TYPE.
    TYPE: str = ""
    # STORE_TYPE is the expected type of store with which AtomicAction is compatible.
    STORE_TYPE: str = ""

    @abc.abstractmethod
    def __init__(self, backend: BaseStoreBackend):
        raise NotImplementedError

    @abc.abstractmethod
    def do(self, keys: Sequence[KeyT], args: Optional[Sequence[StoreValueT]]) -> Any:
        """Execute the AtomicAction on the specified keys with optional arguments.
        :param keys: A sequence of keys.
        :param args: Optional sequence of arguments.
        :return: Any: The result of the AtomicAction.
        """
        raise NotImplementedError

    @classmethod
    def match_or_raise(cls, store_type: str) -> None:
        """Ensure the store type matches the expected type, or raise SetUpError.
        :param store_type: The type of the store to be checked.
        :raise: SetUpError
        """
        if cls.STORE_TYPE != store_type:
            raise SetUpError(
                "Unmatched store type: expected {expected} but got {got}".format(
                    expected=cls.STORE_TYPE, got=store_type
                )
            )


class BaseStore(abc.ABC):
    """Abstract class for all stores."""

    # TYPE is a unique identifier for the type of store.
    TYPE: str = ""

    @classmethod
    def _validate_timeout(cls, timeout: int) -> None:
        """Validate the timeout.
        :param timeout: The timeout in seconds.
        :raise: DataError
        """
        if isinstance(timeout, int) and timeout > 0:
            return

        raise DataError(
            "Invalid timeout: {timeout}, Must be an integer greater "
            "than 0.".format(timeout=timeout)
        )

    @abc.abstractmethod
    def __init__(
        self, server: Optional[str] = None, options: Optional[Dict[str, Any]] = None
    ):
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, key: KeyT) -> bool:
        """Check if the specified key exists.
        :param key: The key to check.
        :return: True if the specified key exists, False otherwise.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def ttl(self, key: KeyT) -> int:
        """Returns the number of seconds until the specified key will expire.
        :param key: The key to check.
        :raise: DataError
        """
        raise NotImplementedError

    @abc.abstractmethod
    def expire(self, key: KeyT, timeout: int) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, key: KeyT, value: StoreValueT, timeout: int) -> None:
        """Set a value for the specified key with specified timeout.
        :param key: The key to set.
        :param value: The value to set.
        :param timeout: The timeout in seconds.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, key: KeyT) -> Optional[StoreValueT]:
        """Get a value for the specified key.
        :param key: The key for which to get a value.
        :return: The value for the specified key, or None if it does not exist.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def hset(
        self,
        name: KeyT,
        key: Optional[KeyT] = None,
        value: Optional[StoreValueT] = None,
        mapping: Optional[StoreDictValueT] = None,
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def hgetall(self, name: KeyT) -> StoreDictValueT:
        raise NotImplementedError

    @abc.abstractmethod
    def make_atomic(self, action_cls: Type[BaseAtomicAction]) -> BaseAtomicAction:
        """Create an instance of an AtomicAction for this store.
        :param action_cls: The class of the AtomicAction.
        :return: The AtomicAction instance.
        """
        raise NotImplementedError



# Copyright (c) Django Software Foundation and individual contributors.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be used
#        to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
from importlib import import_module


def now_sec() -> int:
    return int(time.time())


def now_ms() -> int:
    return int(time.time() * 1000)


def import_string(dotted_path: str):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err
